#include <string>
#include <iostream>
#include <sstream>
#include <stdio.h>
#include <math.h>
#include "ics46goody.hpp"
#include "stopwatch.hpp"

class TN {
  public:
    TN ()                        : left(nullptr), right(nullptr){}
    TN (const TN& tn)            : value(tn.value), left(tn.left), right(tn.right){}
    TN (int v, TN* l = nullptr,
               TN* r = nullptr) : value(v), left(l), right(r){}

    int value;
    TN* left;
    TN* right;
};

//Standard recursive binary tree insertion
void insert (TN*& t, int to_insert) {
  if (t == nullptr)
    t = new TN(to_insert);    //nullptr implicit for left/right subtrees
  else
    if (to_insert < t->value)
      insert(t->left, to_insert);
    else if (to_insert > t->value)
      insert(t->right, to_insert);
}


//Standard iterative binary tree find
TN* find (TN* root, int to_find) {
  for (TN* c=root; c!=nullptr; c = to_find<c->value ? c->left : c->right)
    if (to_find == c->value)
      return c;

  return nullptr;
}


//Standard deallocation
void deallocate(TN* t) {
  if (t == nullptr)
    return;
  else
    deallocate(t->left);
    deallocate(t->right);
    delete t;
}



//Return a std::string with c repeated n times
std::string repeat (char c, int n)
{return ( n<=0 ? "" : c+repeat(c,n-1) );}


//Shuffle the array, putting the values in a random order
//O(N): this isn't hard, but there are lots of "wrong" ways
//  to do it (where it is not completely random)
void shuffle(int* values, int length) {
  for (int i=0; i<length-1; ++i) {
    int index = ics::rand_range(i,length-1);
    int temp  = values[index];
    values[index] = values[i];
    values[i    ] =temp;
  }
}


//Print tree in standard form
void print (TN* t, std::string prefix)  {
  if (t == nullptr)
    return;
  print(t->right,prefix+"  ");
  std::cout << prefix << t->value << std::endl;
  print(t->left,prefix+"  ");
}


//Standard tree height calculation
int height (TN* t)  {
 if (t == nullptr)
    return -1;
  else
    return 1 + std::max(height(t->left),height(t->right));
}


//Returns a weighted height of all nodes in a tree.
//The root is weighted as 1, every node at depth 1 is weighted 2,
//  every node at depth 3 is weighted 3, etc.
//Called as weighted_tree_height(root,1);
long weighted_tree_height (TN* t, long current) {
  if (t == nullptr)
    return 0;
  else
    return current +
        weighted_tree_height(t->left,current+1) + weighted_tree_height(t->right,current+1);
}


//Returns how many values in [0,n-1] are locatable in this tree (should be n)
void contains_check (int on, TN* t, int n) {
  int count = 0;
  for (int i = 0; i<n; i++)
    if (find(t,i) != nullptr)
      count++;
  if (count != n) {//
    std::cout << "Tree " << on << " found only " << count << " of " << n << std::endl;
    print(t,"");//
  }
}


int main() {
  int  size             = ics::prompt_int  ("Enter size of tree to create     ",1000);
  int  num_to_generate  = ics::prompt_int  ("Enter # of random trees to create",100000);
  int  heart_beat       = ics::prompt_int  ("Enter # trees per heartbeat      ",10000);
  bool do_shuffle       = ics::prompt_bool ("Shuffle values before adding     ",true);
  bool verify_search    = ics::prompt_bool ("Verify searches work             ",false);
  bool show_first_tree  = ics::prompt_bool ("Show 1st tree                    ",false);
  int* height_histogram = new int[size];
  int* values           = new int[size];
  long weighted_height  = 0;

  //Randomize seed for random number
  std::srand(std::time(0));

  //Initialize arrays, just once here
  for (int i=0; i<size; i++) {
    values[i]           = i;
    height_histogram[i] = 0;
  }

  std::cout << "\nCollecting Statistics: " << std::endl;
  std::cout << "Processed: " << std::endl;

  ics::Stopwatch t = new ics::Stopwatch();
  t.start();

  //Main loop
  for (int on=0; on<num_to_generate; ++on) {
    //Heart beat
    if ((on+1)%heart_beat == 0)
      std::cout << "  " << on+1 << std::endl;

    //Create permutation
    if (do_shuffle) {
      shuffle(values, size);
    }

    //Build a random tree
    TN* tree = nullptr;
    for (int i=0; i<size; i++)
      insert(tree,values[i]);

    //Show first tree
    if (show_first_tree) {
      std::cout << std::endl;
      print(tree,"");
    }

    //Report if any nodes are not reachable
    if (verify_search)
      contains_check(on,tree,size);

    //Update statistics
    ++height_histogram[height(tree)];
    weighted_height = weighted_height + weighted_tree_height(tree,1);

    //Deallocate tree before building again
    deallocate(tree);
  }
  t.stop();
  std::cout << "  Done" << std::endl;

  //Calculate extremal values
  int sum = 0, minH=size-1, maxH=0, maxI=0;
  for (int i=0; i<size; ++i)
    if (height_histogram[i] != 0) {
      minH = std::min(minH,i);
      maxH = std::max(maxH,i);
      sum += i*height_histogram[i];
      if (height_histogram[i] > height_histogram[maxI])
        maxI = i;
    }

  //Print histogram
  std::cout << "\n\nHeight |   #Trees    | " << std::endl;
  std::cout << repeat('-',70) << std::endl;
  for (int i=minH; i<=maxH; ++i) {
    std::string stars = repeat('*',50*height_histogram[i]/height_histogram[maxI]);
    printf("%6d | %11d | %s\n",
           i, height_histogram[i], stars.c_str());
  }

  //Print all remaining statistics
  std::cout << "\nMinimum Possible Height = " << (log(size+1)/log(2)-1) << std::endl;
  std::cout << "Minimum Actual Height   = "   << minH << std::endl;
  std::cout << "Average Actual Height   = "   << (double)sum/num_to_generate << std::endl;
  std::cout << "Maximum Actual Height   = "   <<maxH << std::endl;
  std::cout << "Maximum Possible Height = "   << +(size-1) << std::endl;


  std::cout << "Avg Compares for Search = " <<
    weighted_height / size / num_to_generate << " (truncated)" << std::endl;
  std::cout << "\nTime = " << t.read() << " seconds; processing speed = " <<
      (int)(num_to_generate/t.read()) << " trees/second" << std::endl;
}
